from typing import Optional

from piou import Cli, Option

cli = Cli(description="A CLI tool")


@cli.command(cmd="foo", help="Run foo command")
def foo_main(
    bar: int = Option(..., help="Bar positional argument (required)"),
    baz: str = Option(..., "-b", "--baz", help="Baz keyword argument (required)"),
    foo: Optional[str] = Option(None, "--foo", help="Foo keyword argument"),
):
    """
    A longer description on what the function is doing.
    You can run it with:
    ```bash
     poetry run python -m piou.example.simple foo 1 -b baz
    ```
    And you are good to go!
    """
    pass


if __name__ == "__main__":
    cli.run()



import asyncio
import dataclasses
import datetime as dt
import inspect
import json
import re
import shlex
from collections import namedtuple
from dataclasses import dataclass, field
from enum import Enum
from inspect import iscoroutinefunction
from pathlib import Path
from typing import (
    Any,
    Optional,
    get_args,
    get_origin,
    get_type_hints,
    Literal,
    TypeVar,
    Generic,
    Callable,
    Union,
    Coroutine,
    cast,
)

from typing_extensions import LiteralString

try:
    from types import UnionType, NoneType  # type: ignore
except ImportError:
    UnionType = Union
    NoneType = type(None)

from uuid import UUID

from .exceptions import (
    PosParamsCountError,
    KeywordParamNotFoundError,
    KeywordParamMissingError,
    InvalidChoiceError,
)


class Password(str):
    pass


class Secret(str):
    pass


T = TypeVar("T", str, int, float, dt.date, dt.datetime, Path, dict, list, Password)


def extract_optional_type(t: Any):
    _origin = get_origin(t)
    if _origin is Union or _origin is UnionType:
        types = tuple(x for x in get_args(t) if x is not NoneType)
        return Union[types]  # type: ignore
    return t


def get_literals_union_args(literal: Any):
    values = []
    if get_origin(literal) is not Literal:
        return values

    for _l in get_args(literal):
        if get_origin(_l) is Union:
            values += get_literals_union_args(_l)
        elif get_origin(_l) is Literal:
            values += get_args(_l)
        else:
            values.append(_l)
    return values


def get_type_hints_derived(f):
    hints = get_type_hints(f)
    fn_parameters = inspect.signature(f).parameters
    _all_hints = {}
    for v in fn_parameters.values():
        _value = hints.get(v.name)
        if _value is None and isinstance(v.default, CommandDerivedOption):
            try:
                _value = get_type_hints(v.default.processor)["return"]
            except KeyError:
                raise ValueError(
                    f"Could not find a return type for attribute {v.name!r}."
                    f"Did you forget to specify the return type of the function?"
                )
        _all_hints[v.name] = _value

    return _all_hints


def validate_value(
    data_type: Any,
    value: str,
    *,
    case_sensitive: bool = True,
    choices: Optional[list[Any]] = None,
    raise_path_does_not_exist: bool = True,
):
    """
    Converts `value` to `data_type`, if not possible raises the appropriate error
    Options:
     - case_sensitive: If True, will not lowercase the value before checking if it's in the choices
     - choices: If set, will check if the value is in the choices
     - raise_path_does_not_exist: If True, will raise a FileNotFoundError if the path does not exist
    """
    _data_type = extract_optional_type(data_type)

    if choices:
        _choices = choices if case_sensitive else [x.lower() for x in choices]
        _value = value if case_sensitive else value.lower()
        if _value not in _choices:
            raise InvalidChoiceError(value, choices)

    if _data_type is Any or _data_type is bool:
        return value
    elif _data_type is str or _data_type is Password:
        return str(value)
    elif _data_type is LiteralString:
        return cast(LiteralString, str(value))
    elif _data_type is int:
        return int(value)
    elif _data_type is float:
        return float(value)
    elif _data_type is UUID:
        return UUID(value)
    elif _data_type is dt.date:
        return dt.date.fromisoformat(value)
    elif _data_type is dt.datetime:
        return dt.datetime.fromisoformat(value)
    elif _data_type is Path:
        p = Path(value)
        if raise_path_does_not_exist and not p.exists():
            raise FileNotFoundError(f'File not found: "{value}"')
        return p
    elif _data_type is dict:
        return json.loads(value)
    elif inspect.isclass(_data_type) and issubclass(_data_type, Enum):
        return _data_type[value].value
    elif _data_type is list or get_origin(_data_type) is list:
        list_type = get_args(_data_type)
        return [validate_value(list_type[0] if list_type else str, x) for x in value.split(" ")]
    elif get_origin(_data_type) is Literal:
        return value
    elif _data_type is list or get_origin(_data_type) is list:
        list_type = get_args(_data_type)
        return [validate_value(list_type[0] if list_type else str, x) for x in value.split(" ")]
    else:
        raise NotImplementedError(f'No parser implemented for data type "{data_type}"')


_KEYWORD_TO_NAME_REG = re.compile(r"^-+")


def keyword_arg_to_name(keyword_arg: str) -> str:
    """Formats a string from '--quiet-v2' to 'quiet_v2'"""
    return _KEYWORD_TO_NAME_REG.sub("", keyword_arg).replace("-", "_")


@dataclass
class CommandOption(Generic[T]):
    default: T
    help: Optional[str] = None
    keyword_args: tuple[str, ...] = field(default_factory=tuple)

    choices: Optional[Union[list[T], Callable[[], list[T]]]] = None

    _name: Optional[str] = field(init=False, default=None)
    _data_type: type[T] = field(init=False, default=Any)  # pyright: ignore[reportAssignmentType]

    # Only for literal types
    case_sensitive: bool = True
    hide_choices: bool = False

    # For dynamic derived
    arg_name: Optional[str] = None

    # Only for Path
    raise_path_does_not_exist: bool = True

    @property
    def data_type(self):
        return self._data_type

    @property
    def literal_values(self):
        return get_literals_union_args(self.data_type)

    @data_type.setter
    def data_type(self, v: type[T]):
        if self.choices and get_literals_union_args(v):
            raise ValueError("Pick either a Literal type or choices")
        self._data_type = v

    @property
    def is_password(self):
        return self.data_type == Password

    @property
    def name(self):
        return self._name or keyword_arg_to_name(sorted(self.keyword_args)[0])

    @name.setter
    def name(self, name: Optional[str]):
        self._name = name

    @property
    def names(self) -> list[str]:
        names = []
        if self._name:
            names.append(self._name)
        for keyword_arg in self.keyword_args:
            names.append(keyword_arg_to_name(keyword_arg))
        return names

    @property
    def is_required(self):
        return self.default is ...

    @property
    def is_positional_arg(self):
        return len(self.keyword_args) == 0

    def get_choices(self) -> Optional[list[T]]:
        _choices = self.choices() if callable(self.choices) else self.choices
        return self.literal_values or _choices

    def validate(self, value: str) -> T:
        _value = validate_value(
            self.data_type,
            value,
            case_sensitive=self.case_sensitive,
            choices=self.get_choices(),
            raise_path_does_not_exist=self.raise_path_does_not_exist,
        )
        return _value  # type: ignore


def Option(
    default: Any,
    *keyword_args: str,
    help: Optional[str] = None,
    # Only for type Literal
    case_sensitive: bool = True,
    arg_name: Optional[str] = None,
    choices: Optional[Any] = None,
    # Only for Path
    raise_path_does_not_exist: bool = True,
) -> Any:
    return CommandOption(
        default=default,
        help=help,
        keyword_args=keyword_args,
        case_sensitive=case_sensitive,
        arg_name=arg_name,
        choices=choices,
        raise_path_does_not_exist=raise_path_does_not_exist,
    )


def _split_cmd(cmd: str) -> list[str]:
    """
    Utility to split a string containing arrays like --foo 1 2 3
    from ['--foo', '1', '2', '3'] to ['--foo', '1 2 3']
    """

    def reset_buff():
        nonlocal buff, cmd_split
        cmd_split.append(" ".join(buff))
        buff = []

    is_pos_arg = True
    buff = []
    cmd_split = []
    for arg in shlex.split(cmd):
        if arg.startswith("-"):
            if buff:
                reset_buff()
            is_pos_arg = False
        else:
            if is_pos_arg:
                cmd_split.append(arg)
                continue
            else:
                buff.append(arg)

        if not buff:
            cmd_split.append(arg)
    if buff:
        reset_buff()
    return cmd_split


def get_cmd_args(cmd: str, types: dict[str, Any]) -> tuple[list[str], dict[str, str]]:
    positional_args = []
    keyword_params = {}

    is_positional_arg = True
    skip_position = None

    cmd_split = _split_cmd(cmd)

    for i, _arg in enumerate(cmd_split):
        if skip_position is not None and i <= skip_position:
            continue

        if _arg.startswith("-"):
            is_positional_arg = False

        if is_positional_arg:
            positional_args.append(_arg)
            continue

        try:
            curr_type = types[keyword_arg_to_name(_arg)]
        except KeyError:
            raise KeywordParamNotFoundError(f"Could not find parameter {_arg!r}", _arg)

        if curr_type is bool:
            keyword_params[_arg] = True
        else:
            keyword_params[_arg] = (
                cmd_split[i + 1]
                if i + 1 < len(cmd_split)
                # In case of "store_true"
                else True
            )
            skip_position = i + 1

    return positional_args, keyword_params


def get_default_args(func) -> list[CommandOption]:
    signature = inspect.signature(func)
    return [v.default for v in signature.parameters.values() if v is not inspect.Parameter.empty]


def parse_input_args(args: tuple[Any, ...], commands: set[str]) -> tuple[Optional[str], list[str], list[str]]:
    """
    Extracts the:
     - global options
     - command
     - command options
     from the passed list or arguments
    """
    global_options, cmd_options, cmd = [], [], None
    for arg in args:
        if cmd is None and arg in commands:
            cmd = arg
            continue

        if cmd is None:
            global_options.append(arg)
        else:
            cmd_options.append(arg)

    if cmd is None and "__main__" in commands:
        cmd = "__main__"
        cmd_options, global_options = global_options, cmd_options

    return cmd, global_options, cmd_options


KeywordParam = namedtuple("KeywordParam", ["name", "validate"])


def convert_args_to_dict(input_args: list[str], options: list[CommandOption]) -> dict:
    _input_pos_args, _input_keyword_args = get_cmd_args(
        " ".join(f"'{x}'" for x in input_args),
        {name: opt.data_type for opt in options for name in opt.names},
    )
    positional_args, keyword_args = [], {}
    for _arg in options:
        if _arg.is_positional_arg:
            positional_args.append(_arg)
        for _keyword_arg in _arg.keyword_args:
            keyword_args[_keyword_arg] = KeywordParam(_arg.arg_name or _arg.name, _arg.validate)

    # Positional arguments
    if len(_input_pos_args) != len(positional_args):
        raise PosParamsCountError(
            f"Expected {len(positional_args)} positional values but got {len(_input_pos_args)}",
            expected_count=len(positional_args),
            count=len(_input_pos_args),
        )

    fn_args = {
        _param.arg_name or _param.name: _param.validate(_param_value)
        for _param, _param_value in zip(positional_args, _input_pos_args)
    }
    # Keyword Arguments
    for _keyword_arg_key, _keyword_arg_value in _input_keyword_args.items():
        _keyword_param = keyword_args.get(_keyword_arg_key)
        if not _keyword_param:
            raise KeywordParamMissingError(
                f"Missing value for required keyword parameter {_keyword_arg_key!r}",
                _keyword_arg_key,
            )
        fn_args[_keyword_param.name] = _keyword_param.validate(_keyword_arg_value)

    # We fill optional fields with None and check for missing ones
    for _arg in options:
        if (_arg.arg_name or _arg.name) not in fn_args:
            if _arg.is_required:
                raise KeywordParamMissingError(
                    f"Missing value for required keyword parameter {_arg.arg_name or _arg.name!r}",
                    _arg.arg_name or _arg.name,
                )
            fn_args[_arg.arg_name or _arg.name] = None if _arg.default is ... else _arg.default

    return fn_args


_LOOP: Optional[asyncio.AbstractEventLoop] = None


def run_function(fn: Callable, *args, **kwargs):
    global _LOOP
    """ Runs an async / non async function """
    if iscoroutinefunction(fn):
        if _LOOP is None:
            try:
                _LOOP = asyncio.get_running_loop()
            except RuntimeError:
                _LOOP = asyncio.new_event_loop()
        return _LOOP.run_until_complete(fn(*args, **kwargs))
    else:
        return fn(*args, **kwargs)


def extract_function_info(
    f,
    from_derived: bool = False,
) -> tuple[list[CommandOption], list["CommandDerivedOption"]]:
    """Extracts the options from a function arguments"""
    options: list[CommandOption] = []
    derived_opts: list[CommandDerivedOption] = []
    for (param_name, param_type), option in zip(get_type_hints_derived(f).items(), get_default_args(f)):
        if isinstance(option, CommandOption):
            # Making a copy in case of reuse
            _option = dataclasses.replace(option)
            if from_derived and _option.arg_name is None:
                # This is to avoid errors when reusing the same parameter name
                # for different derived functions
                _option.arg_name = f"__{f.__name__}.{param_name}"
            _option.name = param_name
            _option.data_type = param_type
            options.append(_option)
        elif isinstance(option, CommandDerivedOption):
            _option = option  # dataclasses.replace(option)
            _option.param_name = param_name
            _options, _ = extract_function_info(_option.processor, from_derived=True)
            options += _options
            derived_opts.append(_option)
        else:
            pass

    return options, derived_opts


@dataclass
class CommandDerivedOption:
    processor: Callable
    """Processor function, can be async """
    param_name: Optional[str] = field(init=False)

    def update_args(self, args: dict) -> dict:
        if self.param_name is None:
            raise ValueError("param_name not set. Did you forget to set it?")
        _args = args.copy()
        fn_args = {}
        _options, _derived = extract_function_info(self.processor, from_derived=True)
        for _opt in _options:
            fn_args[_opt.name] = _args.pop(_opt.arg_name, None) or _args.pop(_opt.name, None)

        for _der in _derived:
            fn_args = _der.update_args(fn_args)

        _args[self.param_name] = run_function(self.processor, **fn_args)
        return _args

    def __repr__(self):
        if hasattr(self, "param_name"):
            return f"<CommandDerivedOption param_name={self.param_name} processor={self.processor}/>"
        else:
            return f"<CommandDerivedOption processor={self.processor}/>"


R = TypeVar("R")


def Derived(processor: Callable[..., Union[Coroutine[Any, Any, R], R]]) -> R:
    return CommandDerivedOption(processor=processor)  # type: ignore



import abc
import shutil
from dataclasses import dataclass
from typing import Optional, Callable

from ..command import Command, CommandOption, ParentArgs, CommandGroup


def get_str_side_by_side(a: str, b: str, col_1_size: int, col_2_size: int, space: int):
    while a or b:
        yield f"{a[:col_1_size].ljust(col_1_size):<{col_1_size + space}}{b[:col_2_size]}"
        a = a[col_1_size:]
        b = b[col_2_size:]


def print_size_by_size(
    print_fn: Callable,
    *args,
    col_1_size: int = 20,
    col_2_size: int = 100,
    space: int = 4,
):
    for arg in args:
        if isinstance(arg, str):
            print_fn(arg)
        else:
            _args = [arg] if isinstance(arg, tuple) else arg
            for x in _args:
                for line in get_str_side_by_side(*x, col_1_size=col_1_size, col_2_size=col_2_size, space=space):
                    print_fn(line)


def get_options_str(
    options: list[CommandOption],
) -> list[tuple[Optional[str], str, CommandOption]]:
    lines = []
    for _option in options:
        if len(_option.keyword_args) == 0:
            name = None
            other_args = None
        else:
            name, *other_args = _option.keyword_args
            other_args = " (" + ", ".join(other_args) + ")" if other_args else ""

        lines.append((name, other_args, _option))
    return lines


@dataclass(frozen=True)
class Titles:
    GLOBAL_OPTIONS = "GLOBAL OPTIONS"
    AVAILABLE_CMDS = "AVAILABLE COMMANDS"
    COMMANDS = "COMMANDS"
    USAGE = "USAGE"
    DESCRIPTION = "DESCRIPTION"
    ARGUMENTS = "ARGUMENTS"
    OPTIONS = "OPTIONS"


@dataclass
class Formatter(abc.ABC):
    print_fn: Callable = print
    col_size: int = 20
    col_space: int = 4

    def print_rows(self, *args):
        columns, _ = shutil.get_terminal_size()
        return print_size_by_size(
            self.print_fn,
            *args,
            col_1_size=self.col_size,
            col_2_size=columns - self.col_size - self.col_space,
            space=self.col_space,
        )

    def print_help(
        self,
        *,
        group: CommandGroup,
        command: Optional[Command] = None,
        parent_args: Optional[ParentArgs] = None,
    ) -> None:
        # We are printing a command help
        if command:
            self.print_cmd_help(command, group.options, parent_args)
        # In case we are printing help for a command group
        elif parent_args:
            self.print_cmd_group_help(group, parent_args)
        # We are printing the CLI help
        else:
            self.print_cli_help(group)

    @abc.abstractmethod
    def print_cli_help(self, group: CommandGroup) -> None:
        ...

    @abc.abstractmethod
    def print_cmd_group_help(self, group: CommandGroup, parent_args: ParentArgs) -> None:
        ...

    @abc.abstractmethod
    def print_cmd_help(
        self,
        command: Command,
        options: list[CommandOption],
        parent_args: Optional[ParentArgs] = None,
    ) -> None:
        ...

    def print_invalid_command(self, available_commands: list[str]) -> None:
        _available_cmds = ", ".join(available_commands)
        self.print_fn(f'Unknown command given. Possible commands are "{_available_cmds}"')

    def print_param_error(self, key: str, cmd: str) -> None:
        self.print_fn(f"Could not find value for {key!r} in command {cmd!r}")

    def print_keyword_param_error(self, cmd: str, param: str) -> None:
        self.print_fn(f"Could not find keyword parameter {param!r} for command {cmd!r}")

    def print_count_error(self, expected_count: int, count: int, cmd: str) -> None:
        self.print_fn(f"Expected {expected_count} positional arguments but got {count} for command {cmd!r}")

    def print_invalid_value_error(self, value: str, choices: list[str]) -> None:
        possible_fields = "\n" + "\n - ".join(_choice for _choice in choices)
        self.print_fn(f"Invalid value {value!r} found. Possible values are: {possible_fields}")

    def print_error(self, message: str) -> None:
        self.print_fn(message)



from .cli import Cli, CommandGroup
from .utils import Option, Derived, Password
from .command import CommandMeta
from .exceptions import CommandNotFoundError, CommandError

__all__ = (
    "Cli",
    "CommandGroup",
    "Option",
    "Derived",
    "Password",
    "CommandMeta",
    "CommandNotFoundError",
    "CommandError",
)



import sys
from dataclasses import dataclass, field
from typing import Optional, Union

from rich.console import Console, RenderableType
from rich.markdown import Markdown
from rich.padding import Padding
from rich.table import Table

from .base import Formatter, Titles
from ..command import Command, CommandOption, ParentArgs, CommandGroup


def pad(s: RenderableType, padding_left: int = 1):
    return Padding(s, (0, padding_left))


def fmt_option(option: CommandOption, show_full: bool = False, color: str = "white") -> str:
    if option.is_positional_arg:
        return f"[{color}]<{option.name}>[/{color}]"
    elif show_full:
        first_arg, *other_args = option.keyword_args
        required = f"[{color}]*[/{color}]" if option.is_required else ""
        if other_args:
            other_args = ", ".join(other_args)
            return f"[{color}]{first_arg}[/{color}] ({other_args}){required}"
        else:
            return f"[{color}]{first_arg}[/{color}]{required}"
    else:
        return "[" + sorted(option.keyword_args)[-1] + "]"


def fmt_cmd_options(options: list[CommandOption]) -> str:
    return (
        " ".join([fmt_option(x) for x in options]) if options else ""  # '[<arg1>] ... [<argN>]'
    )


def fmt_help(
    option: CommandOption,
    show_default: bool,
    *,
    markdown_open: Optional[str] = "[bold]",
    markdown_close: Optional[str] = "[/bold]",
):
    _choices = option.get_choices()
    _markdown_open, _markdown_close = markdown_open or "", markdown_close or ""

    if show_default and option.default is not None and not option.is_required:
        default_str = option.default if not option.is_password else "******"
        default_str = f"{_markdown_open}(default: {default_str}){_markdown_close}"
        return option.help + f" {default_str}" if option.help else default_str
    elif _choices is not None and not option.hide_choices:
        if len(_choices) <= 3:
            possible_choices = ", ".join(str(_choice) for _choice in _choices)
            choices_help = f"{_markdown_open}(choices are: {possible_choices}){_markdown_close}"
        else:
            sep = " \n - "
            possible_choices = sep + sep.join(str(_choice) for _choice in _choices)
            choices_help = f"\n{_markdown_open}Possible choices are:" + possible_choices + _markdown_close
        return option.help + f" {choices_help}" if option.help else choices_help
    else:
        return option.help


def get_usage(
    global_options: list[CommandOption],
    command: Optional[str] = None,
    command_options: Optional[list[CommandOption]] = None,
    parent_args: Optional[ParentArgs] = None,
):
    parent_args = parent_args or []
    _global_options = " ".join(["[" + sorted(x.keyword_args)[-1] + "]" for x in global_options])
    _command = None
    if command != "__main__":
        _command = f"[underline]{command}[/underline]" if command else "<command>"
    cmds = [sys.argv[0].split("/")[-1]] + [x.cmd for x in parent_args]
    cmds = " ".join(f"[underline]{x}[/underline]" for x in cmds)

    usage = cmds

    if _global_options:
        usage = f"{usage} {_global_options}"
    usage = f"{usage} {_command}" if _command is not None else usage
    if command_options:
        usage = f"{usage} {fmt_cmd_options(command_options)}"

    return usage


@dataclass(frozen=True)
class RichTitles(Titles):
    GLOBAL_OPTIONS = f"[bold]{Titles.GLOBAL_OPTIONS}[/bold]"
    AVAILABLE_CMDS = f"[bold]{Titles.AVAILABLE_CMDS}[/bold]"
    COMMANDS = f"[bold]{Titles.COMMANDS}[/bold]"
    USAGE = f"[bold]{Titles.USAGE}[/bold]"
    DESCRIPTION = f"[bold]{Titles.DESCRIPTION}[/bold]"
    ARGUMENTS = f"[bold]{Titles.ARGUMENTS}[/bold]"
    OPTIONS = f"[bold]{Titles.OPTIONS}[/bold]"


MIN_MARKDOWN_SIZE: int = 75


@dataclass
class RichFormatter(Formatter):
    _console: Console = field(init=False, default_factory=lambda: Console(markup=True, highlight=False))
    cmd_color: str = "cyan"
    option_color: str = "cyan"
    default_color: str = "white"
    show_default: bool = True
    """Use Markdown object for the description, otherwise use 
    default str
    """
    use_markdown: bool = True
    """See https://pygments.org/styles/ for a list of styles """
    # Only usable if use_markdown is True
    code_theme: str = "solarized-dark"

    def _color_cmd(self, cmd: str):
        return f"[{self.cmd_color}]{cmd}[/{self.cmd_color}]"

    def __post_init__(self):
        self.print_fn = self._console.print

    def _print_description(self, item: Union[CommandGroup, Command]):
        description = item.description or item.help
        if description:
            self.print_fn()
            self.print_fn(RichTitles.DESCRIPTION)
            if self.use_markdown:
                _max_width = max(len(x) for x in description.split("\n"))
                self.print_fn(
                    pad(
                        Markdown(
                            "  \n".join(description.split("\n")),
                            code_theme=self.code_theme,
                        )
                    ),
                    width=max(_max_width, MIN_MARKDOWN_SIZE),
                )
            else:
                self.print_fn(pad(description))

    def _fmt_help(self, option: CommandOption):
        return fmt_help(
            option,
            self.show_default,
            markdown_open=f"[{self.default_color}][bold]",
            markdown_close=f"[/{self.default_color}][/bold]",
        )

    def _print_options(self, options: list[CommandOption]):
        self.print_rows(
            [
                (
                    fmt_option(opt, show_full=True, color=self.option_color),
                    self._fmt_help(opt),
                )
                for opt in options
            ]
        )

    def print_rows(self, rows: list[tuple[str, Optional[str]]]):  # pyright: ignore[reportIncompatibleMethodOverride]
        table = Table(show_header=False, box=None, padding=(0, self.col_space))
        table.add_column(width=self.col_size)
        table.add_column()
        for row in rows:
            table.add_row(*row)
        self.print_fn(table)

    def print_cli_help(self, group: CommandGroup):
        self.print_fn(RichTitles.USAGE)
        self.print_fn(pad(get_usage(group.options)))
        self.print_fn()

        if group.options:
            self.print_fn(RichTitles.GLOBAL_OPTIONS)
            self._print_options(group.options)
            self.print_fn()

        self.print_fn(RichTitles.AVAILABLE_CMDS)
        self.print_rows(
            [(f' {self._color_cmd(_command.name or "")}', _command.help) for _command in group.commands.values()]
        )
        self._print_description(group)

    def print_cmd_help(
        self,
        command: Command,
        options: list[CommandOption],
        parent_args: Optional[ParentArgs] = None,
    ):
        usage = get_usage(
            global_options=options,
            command=command.name,
            command_options=command.options_sorted,
            parent_args=parent_args,
        )
        self.print_fn(RichTitles.USAGE)
        self.print_fn(pad(usage))
        self.print_fn()

        if command.positional_args:
            self.print_fn(RichTitles.ARGUMENTS)
            self.print_rows(
                [
                    (
                        fmt_option(option, color=self.option_color),
                        self._fmt_help(option),
                    )
                    for option in command.positional_args
                ]
            )
        if command.keyword_args:
            self.print_fn("\n" + RichTitles.OPTIONS)
            self._print_options(command.keyword_args)

        global_options = options + [
            parent_option for parent_arg in (parent_args or []) for parent_option in parent_arg.options
        ]
        if global_options:
            self.print_fn("\n" + RichTitles.GLOBAL_OPTIONS)
            self._print_options(global_options)

        self._print_description(command)

    def print_cmd_group_help(self, group: CommandGroup, parent_args: ParentArgs):
        parent_commands = [sys.argv[0].split("/")[-1]] + [x.cmd for x in parent_args]
        commands_str = []
        for i, (cmd_name, cmd) in enumerate(group.commands.items()):
            _cmds = []
            for cmd_lvl, x in enumerate(parent_commands + [cmd_name]):
                _cmds.append(f"[underline]{x}[/underline]")
                if group.options and cmd_lvl == len(parent_commands) - 1:
                    _cmds.append(fmt_cmd_options(group.options))
            _cmds_str = " ".join(_cmds)
            _line = f'{"" if i == 0 else "or: ":>5}{_cmds_str} {fmt_cmd_options(cmd.options_sorted)}'.rstrip()
            commands_str.append(_line)
        commands_str = "\n".join(commands_str)

        self.print_fn(RichTitles.USAGE)
        self.print_fn(commands_str)

        self.print_fn()

        self.print_fn(RichTitles.COMMANDS)
        for cmd_name, cmd in group.commands.items():
            self.print_fn(pad(f"[underline]{cmd_name}[/underline]", padding_left=2))
            if cmd.help:
                self.print_fn(pad(cmd.help, padding_left=4))
                self.print_fn()
            if cmd.options:
                self.print_rows(
                    [
                        (
                            fmt_option(opt, show_full=True, color=self.option_color),
                            self._fmt_help(opt),
                        )
                        for opt in cmd.options_sorted
                    ]
                )
                self.print_fn()

        if group.options:
            self.print_fn(RichTitles.OPTIONS)
            self._print_options(group.options)
            self.print_fn()

        global_options = [parent_option for parent_arg in (parent_args or []) for parent_option in parent_arg.options]
        if global_options:
            self.print_fn(RichTitles.GLOBAL_OPTIONS)
            self._print_options(global_options)

        self._print_description(group)

    def print_error(self, message: str):
        self.print_fn(f"[red]{message}[/red]")

    def print_invalid_command(self, available_commands: list[str]):
        _available_cmds = ", ".join(available_commands)
        self.print_error(f'Unknown command given. Possible commands are "[bold]{_available_cmds}[/bold]"')

    def print_keyword_param_error(self, cmd: str, param: str) -> None:
        self.print_error(f"Could not find keyword parameter [bold]{param!r}[/bold] for command [bold]{cmd!r}[/bold]")

    def print_param_error(self, key: str, cmd: str) -> None:
        self.print_error(f"Could not find value for [bold]{key!r}[/bold] in [bold]{cmd}[/bold]")

    def print_count_error(self, expected_count: int, count: int, cmd: str):
        self.print_error(
            f"Expected {expected_count} positional arguments but got {count} for command [bold]{cmd}[/bold]"
        )

    def print_invalid_value_error(self, value: str, choices: list[str]):
        sep = "\n - "
        possible_fields = sep + sep.join(_choice for _choice in choices)
        msg = f"Invalid value [bold]{value}[/bold] found. Possible values are: {possible_fields}"
        self.print_error(msg)



import textwrap
from dataclasses import dataclass, field
from functools import wraps
from inspect import getdoc
from typing import Optional, Any, NamedTuple, Callable

from .exceptions import DuplicatedCommandError, CommandException, CommandNotFoundError
from .utils import (
    CommandOption,
    CommandDerivedOption,
    extract_function_info,
    parse_input_args,
    run_function,
    convert_args_to_dict,
)


class ParentArg(NamedTuple):
    cmd: str
    options: list[CommandOption]
    input_options: list[str]
    options_processor: Optional[Callable] = None
    propagate_args: Optional[bool] = False


ParentArgs = list[ParentArg]


def clean_multiline(s: str) -> str:
    return textwrap.dedent(s).strip()


@dataclass
class Command:
    name: str
    fn: Callable
    help: Optional[str] = None
    options: list[CommandOption] = field(default_factory=list)
    description: Optional[str] = None
    derived_options: list[CommandDerivedOption] = field(default_factory=list)

    @property
    def positional_args(self) -> list[CommandOption]:
        return [opt for opt in self.options if opt.is_positional_arg]

    @property
    def keyword_args(self) -> list[CommandOption]:
        return sorted(
            [opt for opt in self.options if not opt.is_positional_arg],
            key=lambda x: (-x.is_required, x.name),
        )

    @property
    def options_sorted(self) -> list[CommandOption]:
        """Sorts with the following order:
        - positional
        - keyword required
        - keyword optional
        """
        return self.positional_args + self.keyword_args

    def run(self, *args, **kwargs):
        run_function(self.fn, *args, **kwargs)

    def __post_init__(self):
        self.description = clean_multiline(self.description) if self.description else None

        keyword_params = [x for x in self.options if not x.is_positional_arg]
        if not keyword_params:
            return

        _keyword_args = set()
        for _param in keyword_params:
            for _keyword_arg in _param.keyword_args:
                if _keyword_arg in _keyword_args:
                    raise ValueError(f'Duplicate keyword args found "{_keyword_arg}"')
                _keyword_args.add(_keyword_arg)


@dataclass
class CommandMeta:
    cmd_name: str
    fn_args: dict
    cmd_args: dict


OnCommandRun = Callable[[CommandMeta], None]


@dataclass
class CommandGroup:
    name: Optional[str] = None
    help: Optional[str] = None
    """ Short line to explain the command group"""
    description: Optional[str] = None
    """ Description of the command group"""

    options_processor: Optional[Callable] = None

    propagate_options: Optional[bool] = None

    on_cmd_run: Optional[OnCommandRun] = None

    # _formatter: Formatter = field(init=False, default=None)
    _options: list[CommandOption] = field(init=False, default_factory=list)
    _commands: dict[str, Command] = field(init=False, default_factory=dict)
    _command_groups: dict[str, "CommandGroup"] = field(init=False, default_factory=dict)

    #

    def __post_init__(self):
        self.description = clean_multiline(self.description) if self.description else None

    @property
    def commands(self) -> dict:
        return {k: self._commands.get(k) or self._command_groups[k] for k in sorted(self.command_names)}

    @property
    def options(self):
        return sorted(self._options, key=lambda x: x.is_required, reverse=True)

    def add_sub_parser(self, help: Optional[str] = None, description: Optional[str] = None):
        cls = type(self)
        return cls(help=help, description=description)  # noqa

    def add_option(
        self,
        *args: str,
        help: Optional[str] = None,
        data_type: Any = bool,
        default: Any = False,
    ):
        opt = CommandOption(
            default=default,
            help=help,
            keyword_args=args,
        )
        opt.data_type = data_type
        self._options.append(opt)

    @property
    def command_names(self) -> set[str]:
        return set(self._commands.keys()) | (self._command_groups.keys())

    def add_group(self, group: "CommandGroup"):
        if group.name is None:
            raise NotImplementedError("A group must have a name")

        if group.name in self.command_names:
            raise DuplicatedCommandError(f"Duplicated command found for {group.name!r}", group.name)
        group.on_cmd_run = self.on_cmd_run
        self._command_groups[group.name] = group

    def add_command(
        self,
        f,
        cmd: Optional[str] = None,
        help: Optional[str] = None,
        description: Optional[str] = None,
        is_main: bool = False,
    ):
        cmd_name = "__main__" if is_main else cmd or f.__name__
        if cmd_name in self.commands:
            raise DuplicatedCommandError(f"Duplicated command found for {cmd_name!r}", cmd_name)

        if cmd_name == "__main__" and self.commands:
            raise CommandException("Main command cannot be added with other commands")
        if cmd_name != "__main__" and "__main__" in self.commands:
            raise CommandException(f"Command {cmd_name!r} cannot be added with main command")

        _options, _derived_options = extract_function_info(f)
        self._commands[cmd_name] = Command(
            name=cmd_name,
            fn=f,
            options=_options,
            derived_options=_derived_options,
            help=help,
            description=description or getdoc(f),
        )

    def command(
        self,
        cmd: Optional[str] = None,
        help: Optional[str] = None,
        description: Optional[str] = None,
        is_main: bool = False,
    ):
        def _command(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            self.add_command(f, cmd=cmd, help=help, description=description, is_main=is_main)
            return wrapper

        return _command

    def processor(self):
        def _processor(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            options, _ = extract_function_info(f)
            for option in options:
                self._options.append(option)
            self.set_options_processor(f)
            return wrapper

        return _processor

    def run_with_args(self, *args, parent_args: Optional[ParentArgs] = None):
        cmd, global_options, cmd_options = parse_input_args(args, self.command_names)
        command_group = self._command_groups.get(cmd) if cmd else None
        command = (command_group or self)._commands.get(cmd) if cmd else None

        parent_args = parent_args or []
        if command_group:
            if cmd is None:
                raise NotImplementedError('"cmd" cannot be empty')
            parent_args.append(
                ParentArg(
                    cmd,
                    self.options,
                    global_options,
                    options_processor=self.options_processor,
                    propagate_args=self.propagate_options,
                )
            )
            return command_group.run_with_args(*cmd_options, parent_args=parent_args)

        if set(global_options + cmd_options) & {"-h", "--help"}:
            raise ShowHelpError(group=command_group or self, parent_args=parent_args, command=command)

        if not command:
            raise CommandNotFoundError(list(self.command_names))

        args_dict = {}
        options, input_options, processors, propagate_args = (
            # Parent / current
            [command.options, self._options],
            [cmd_options, global_options],
            [None, self.options_processor],
            [True, self.propagate_options],
        )

        for parent_arg in parent_args:
            options.append(parent_arg.options)
            input_options.append(parent_arg.input_options)
            processors.append(parent_arg.options_processor)
            propagate_args.append(parent_arg.propagate_args)

        for _opts, _input_opts, _processor, _propagate in zip(options, input_options, processors, propagate_args):
            try:
                _arg_dict = convert_args_to_dict(_input_opts, _opts)
            except CommandException as e:
                e.cmd = cmd
                raise e
            if _processor:
                _processor(**_arg_dict)
            if _propagate is not False:
                args_dict.update(_arg_dict)

        cmd_args = args_dict.copy()
        for _derived in command.derived_options:
            args_dict = _derived.update_args(args_dict)

        if self.on_cmd_run:
            full_command_name = ".".join([x.cmd for x in parent_args] + [command.name])
            self.on_cmd_run(CommandMeta(full_command_name, fn_args=args_dict, cmd_args=cmd_args))

        return command.run(**args_dict)

    def set_options_processor(self, fn: Callable):
        self.options_processor = fn
        # We don't propagate if not specified otherwise when processor is set
        if self.propagate_options is None:
            self.propagate_options = False


class ShowHelpError(Exception):
    def __init__(
        self,
        group: CommandGroup,
        command: Optional[Command] = None,
        parent_args: Optional[ParentArgs] = None,
    ):
        self.command = command
        self.group = group
        self.parent_args = parent_args
